/* Sensoray 518 Driver
 *
 * Copyright (C) 2019  Sensoray Co., Inc.
 * GPL
 * @author: D.A.
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
#include <linux/config.h>
#endif
#include <linux/module.h>
#include <linux/cdev.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/wait.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/pagemap.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 17)
#include <linux/uaccess.h>
#include <linux/io.h>
#endif
#include <linux/irq.h>
#include <linux/spinlock.h>
#include <linux/moduleparam.h>
#include <linux/kthread.h>

#include "s518ioctl.h"

MODULE_LICENSE("GPL");

#define S518_VERSION2	1
#define S518_VERSION1	0
#define S518_VERSION0	2
#define S518_VERSION		KERNEL_VERSION(S518_VERSION2,	\
					       S518_VERSION1,	\
					       S518_VERSION0)
#define S518_VERSION_TEXT "Sensoray 518 driver, version 1.0.2"

/* for backward compatibility */
#define KERNEL_VERSION_NOINODE KERNEL_VERSION(2, 6, 29)
#define KERNEL_VERSION_IOCTL_REMOVED KERNEL_VERSION(2, 6, 36)
#define KERNEL_VERSION_MUTEX_ADDED KERNEL_VERSION(2, 6, 16)
#if LINUX_VERSION_CODE >= KERNEL_VERSION_MUTEX_ADDED
#define s518_mutex_init mutex_init
#define s518_mutex_lock mutex_lock
#define s518_mutex_lock_interruptible mutex_lock_interruptible
#define s518_mutex_unlock mutex_unlock
#else
#define s518_mutex_init init_MUTEX
#define s518_mutex_lock down
#define s518_mutex_lock_interruptible down_interruptible
#define s518_mutex_unlock up
#endif

#define S518_MAJOR 127
#define ADDRESS_SPACE 4

#define MAX_HS_TIME 1000 // 1000 milliseconds

	static int debug;
static int *s518_debug = &debug;

static unsigned int s518_major = -1;
static struct class *s518_class;
static int major;
module_param(debug, int, 0);
MODULE_PARM_DESC(debug, "Debug level(0-100) default 0");
int num_cards = 0;

#define MAX_BOARDS 16
#define MAX_CARDS MAX_BOARDS

static unsigned long io[MAX_BOARDS] = {[0 ... (MAX_BOARDS-1)] = 0};
static int s512devs[2];

module_param(major, int, 0444);
module_param_array(io, ulong, &s512devs[0], 0444);
MODULE_PARM_DESC(major, "Driver major number (default="
		 __MODULE_STRING(DFLT_MAJOR) ")");
MODULE_PARM_DESC(io, "Base I/O address - not for PCI");

#define dprintk(level, fmt, arg...)				\
	do {							\
		if (*s518_debug >= (level)) {			\
			printk(KERN_DEBUG "s518: " fmt, ##arg);	\
		}						\
	} while (0)

#define NUM_518_CHANNELS 8
#define DEFAULT_518_LSDC 2

struct s518_channel {
	int lsdc;           // logical sensor definition code
	short HiLim;	    //  High alarm limit.
	short LoLim;		//  Low alarm limit.
	int	LoLimEnable;	//  High alarm limit enabled.
	int	HiLimEnable;	//  Low alarm limit enabled.
	int	LoAlarm;		//  High alarm detected.
	int	HiAlarm;		//  Low alarm detected.
	int	FailHigh;		//  Fail high (versus low) on open sensor detected.
	char Filter;		//  Filter factor.
	short GagemVperV;	//	Gage mV/V * 10 (STDbus products only).
	short GageMaxLoad;  //	Data value expected at max gage load (STDbus only).
};

struct s518_dev {
	int board_state;
	int is_mode_reject_50hz;
	int is_mode_high_speed;
	int fault_flags;
	int io;
	struct s518_channel ch[NUM_518_CHANNELS];
	int users;
	int minor;
#if LINUX_VERSION_CODE >= KERNEL_VERSION_MUTEX_ADDED
	struct mutex open_lock;
#else
	struct semaphore open_lock;
#endif	
	int in_reset;
#if LINUX_VERSION_CODE >= KERNEL_VERSION_MUTEX_ADDED
	struct mutex reset_lock;
#else
	struct semaphore reset_lock;
#endif	
	
};


struct s518_fh {
	struct s518_dev *dev;
	int io;
};

struct s518_dev s518_device[MAX_BOARDS];

#define DIAG_GETRAWDATA		1		// Return channel's raw data.
#define DIAG_GETEEPROM      2
#define DIAG_GETPRODUCTID	4		// Return product model number.
#define DIAG_GETVERSION		5		// Return firmware version number.
#define DIAG_HISPEEDMODE	8		// Switch to high-speed mode.

#define CMD_RDCHAN			 0		// Get data from 1 channel
#define CMD_DFNTYP			 1		// Declare sensor type
#define CMD_DFNLIM			 2		// Set alarm limits
#define CMD_RDALRM			 3		// Get alarm flags for 8 channels
#define CMD_RDTEMP			 4		// Get TB temperature(s)
#define CMD_DFNOPN			 5		// Set open-sensor values
#define CMD_TARE			 6		// Tare gage
#define CMD_FILT50			 7		// Set 50Hz filter mode
#define CMD_READ8			 8		// Get data from 8 channels
#define CMD_CAL				10		// Calibration function
#define CMD_DIAGNOSE		11		// Diagnostic function
#define CMD_RDSLOP			12		// Get gage slope/intercept
#define CMD_WTSLOP			13		// Set gage slope/intercept
#define CMD_GZERO			14		// Set gage zero
#define CMD_GSPAN			15		// Set gage span
#define CMD_STFILT			16		// Set filter factor
#define CMD_SLEEP			17		// Switch to low-power standby
#define CMD_WAKEUP			18		// Wake from low-power standby
#define CMD_WCOEFF			19		// Set custom-R coefficients

static const char Cmd518[] = {
	(const char)0x00,	// CMD_RDCHAN	--	Get data from channel
	(const char)0x10,	// CMD_DFNTYP	--	Declare sensor type
	(const char)0x20,	// CMD_DFNLIM	--	Set alarm limits
	(const char)0x30,	// CMD_RDALRM	--	Get alarm flags for 8 channels
	(const char)0x40,	// CMD_RDTEMP	--	Get TB temperature(s)
	(const char)0x50,	// CMD_DFNOPN	--	Set open-sensor values
	(const char)0x70,	// CMD_TARE		--	Tare gage
	(const char)0x48,	// CMD_FILT50	--	Set 50Hz filter mode
	(const char)0x58,	// CMD_READ8	--	Get data from 8 channels
	(const char)0xFF,	//					 spare
	(const char)0xE0,	// CMD_CAL		--	Calibration function
	(const char)0xF0,	// CMD_DIAGNOSE	--	Diagnostic function
	(const char)0x80,	// CMD_RDSLOP	--	Get gage slope/intercept
	(const char)0x90,	// CMD_WTSLOP	--	Set gage slope/intercept
	(const char)0xB0,	// CMD_GZERO	--	Set gage zero
	(const char)0xD0,	// CMD_GSPAN	--	Set gage span
	(const char)0x60,	// CMD_STFILT	--	Set filter factor
	(const char)0x43,	// CMD_SLEEP	--	Switch to low-power standby
	(const char)0x42,	// CMD_WAKEUP	--	Wake from low-power standby
	(const char)0xC0	// CMD_WCOEFF	--	Set custom-R coefficients
};

// Logical-to-physical SDC mapping table.  "**" = Legacy types.
// PhysSDC
static const char Sdc518[] = {
	(const char)0x00,	//  0 : 10V **
	(const char)0x21,	//  1 : 10V
	(const char)0x00,	//  2 : 5V **
	(const char)0x15,	//  3 : 5V
	(const char)0x0E,	//  4 : 1.65V **
	(const char)0x16,	//  5 : 500mV
	(const char)0x17,	//  6 : 100mV
	(const char)0x0D,	//  7 : 80mV **
	(const char)0x01,	//  8 : E
	(const char)0x02,	//  9 : J **
	(const char)0x1B,	// 10 : J
	(const char)0x03,	// 11 : K **
	(const char)0x1C,	// 12 : K
	(const char)0x04,	// 13 : T **
	(const char)0x1D,	// 14 : T
	(const char)0x05,	// 15 : S **
	(const char)0x1E,	// 16 : S
	(const char)0x06,	// 17 : R **
	(const char)0x1F,	// 18 : R
	(const char)0x07,	// 19 : 385 **
	(const char)0x18,	// 20 : 385 platinum
	(const char)0x08,	// 21 : 392 **
	(const char)0x19,	// 22 : 392 platinum
	(const char)0x0B,	// 23 : 10K **
	(const char)0x1A,	// 24 : 10K
	(const char)0x11,	// 25 : 4-20 mA
	(const char)0x0F,	// 26 : Strain/pressure gage
	(const char)0x0A,	// 27 : 400 ohm range
	(const char)0x14,	// 28 : 4K ohm range
	(const char)0x20,	// 29 : 600K ohm range
	(const char)0x0C,	// 30 : Custom resistive sensor
	(const char)0x13,	// 31 : Disabled: remove from scan list
	(const char)0x22,	// 32 : N
	(const char)0x23,	// 33 : C
	(const char)0x24,	// 34 : B
	(const char)0xFF,	// 35 : 120 nickel RTD (STDbus only)
	(const char)0x28,	// 36 : 200 nickel RTD
	(const char)0x29,	// 37 : 1K nickel RTD
	(const char)0x2A,	// 38 : 385 platinum, HiRes (0.0125C)
	(const char)0x2B,	// 39 : 392 platinum, HiRes (0.0125C)
	(const char)0x25,	// 40 : 10K thermistor on fixed 600K ohm range
	(const char)0x2C,	// 41 : 10 ohm copper RTD
	(const char)0x14,	// 42 : 1K nickel RTD (4Kohm range, ext lin)
	(const char)0x14	// 43 : 1K platinum RTD (4Kohm range, ext lin)
};


void s518_outb(struct s518_dev *dev, unsigned char val, unsigned short int port, int throttle)
{
	outb(val, port);
	return;
}

unsigned char s518_inb(struct s518_dev *dev, unsigned short int port)
{
	unsigned char val;
	val = inb(port);
	return val;
}

int timespec_2_ms(struct timespec *t)
{
	int ms;
	ms = t->tv_nsec / 1000000 + t->tv_sec * 1000;
	return ms;
}

int timespec_subtract(struct timespec *result, struct timespec *x, struct timespec *y)
{
	/* Perform the carry for the later subtraction by updating y. */
	if (x->tv_nsec < y->tv_nsec) {
		int nsec = (y->tv_nsec - x->tv_nsec) / 1000000000 + 1;
		y->tv_nsec -= 1000000000 * nsec;
		y->tv_sec += nsec;
	}
	if (x->tv_nsec - y->tv_nsec > 1000000000) {
		int nsec = (x->tv_nsec - y->tv_nsec) / 1000000000;
		y->tv_nsec += 1000000000 * nsec;
		y->tv_sec -= nsec;
	}

	/* Compute the time remaining to wait.
	   tv_usec is certainly positive. */
	result->tv_sec = x->tv_sec - y->tv_sec;
	result->tv_nsec = x->tv_nsec - y->tv_nsec;

	/* Return 1 if result is negative. */
	return x->tv_sec < y->tv_sec;
}


int handshakewait(struct s518_dev *dev, char mask, char xormask, int mstime)
{
	struct timespec t1;//start time
	struct timespec t2;
	struct timespec elapsed;
	unsigned char val;
	int j = 5;
	//clock_gettime(CLOCK_MONOTONIC, &t1);
	getnstimeofday(&t1);
	do {
		val = s518_inb(dev, dev->io + 1);
		if (val & 0x10) {
			if (!dev->in_reset) {
				printk("518 firmware fault.  board reset required!\n");
				dev->fault_flags |= 0x10;
			}
           
		}

		if (((val & mask) ^ xormask ) != 0) {
			dprintk(2, "passed\n");
			return 0;
		}
		//clock_gettime(CLOCK_MONOTONIC, &t2);
		getnstimeofday(&t2);        
		timespec_subtract(&elapsed, &t1, &t2);

	} while (abs(timespec_2_ms(&elapsed)) < abs(mstime));

	if (((val & mask) ^ xormask ) != 0) {
		printk("passed\n");
		return 0;
	}
	printk("handshake wait\n");
	return -EAGAIN;
}

int send_command(struct s518_dev *dev, char *cmd, int size)
{
	int i;
	int rc;
	if (size > MAX_CMD_BYTES)
		return -EINVAL;
        
	for (i = 0; i < size; i++) {
		rc = handshakewait(dev, STAT_CRMT, 0, MAX_HS_TIME);
		if (rc != 0) {
			printk("send_command failed %x\n", rc);
			return -EAGAIN;
		}
		// short delay
		(void) handshakewait(dev, STAT_CRMT, 0, 0);
		s518_outb(dev, cmd[i], dev->io, i == 0);
	}
	// save the sensor command.
	// doing this in the library doesn't save the state between
	// program instances.  It belongs here in the driver.
	if (((cmd[0] & 0xf0) == 0x10) && (size > 1)) {
		struct s518_channel *ch;
		int idx;
		int j;
		idx = cmd[0] & 0x07;
		// define sensor
		ch = &dev->ch[idx];
		for (j = 0; j < 44; j++) {
			if (Sdc518[j] == cmd[1])
				break;
		}
		if (j == 44)
			j = 0;
        
		printk("setting log SDC to %x for channel %d, %p\n", j, idx, ch);
		ch->lsdc = j;
		// changing sensor type requires wait or data will be corrupt when read back
		// there is no handshake on the board
		msleep(10); 
	} else if (cmd[0] == Cmd518[CMD_DIAGNOSE] && cmd[1] == DIAG_HISPEEDMODE) {
		dev->is_mode_high_speed = 1;
		printk("setting high speed mode\n");
	} else if (cmd[0] == Cmd518[CMD_FILT50]) {
		dev->is_mode_reject_50hz = 1;
		printk("setting reject 50hz mode\n");
	}
	return 0;
}

static int init_channel(struct s518_channel *ch, short lsdc)
{
	if (ch == NULL)
		return -EINVAL;
	if (lsdc < 0)
		return -EINVAL;
	ch->lsdc = lsdc;
	ch->HiLim = 32767;
	ch->LoLim = -32768;
	ch->Filter = 0;
	ch->FailHigh = 1;
	ch->LoLimEnable = 0;
	ch->HiLimEnable = 0;
	ch->LoAlarm = 0;
	ch->HiAlarm = 0;
	return 0;
}

static  void init_all_channels(struct s518_dev *dev)
{
	int i;
	for (i = 0; i < NUM_518_CHANNELS; i++) {
		init_channel(&dev->ch[i], DEFAULT_518_LSDC);
	}
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION_NOINODE
static long s518_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
#else
	static int sx16_ioctl(struct inode *inode, struct file *file,
			      unsigned int cmd, unsigned long arg)
#endif

{
	unsigned long val = (unsigned long) arg;
	/* Board number.*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION_NOINODE
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 19, 0)
	struct inode *inode = file->f_dentry->d_inode;
#else
	struct inode *inode = file_inode(file);
#endif
#endif
	int board = MINOR(inode->i_rdev);
	struct s518_fh *fh = (struct s518_fh *) file->private_data;
	struct s518_dev *dev = (struct s518_dev *) fh->dev;	
	void __user *argp = (void __user *)arg;
	int rc = 0;
	dprintk(8, "%s, %d\n", __func__, cmd);
	switch (cmd) {
	case S518_IOC_VERSION:
	{
		struct s518_version p;
		p.ver = S518_VERSION2 << 16;
		p.ver |= S518_VERSION1 << 8;
		p.ver |= S518_VERSION0;
		rc = copy_to_user(argp, &p, sizeof(struct s518_version));
		dprintk(4, "s518 ioc version\n");
		break;
	}
	case S518_IOC_WRITE_BYTE:
	{
		struct s518_cmd p;
		rc = copy_from_user(&p, argp, sizeof(struct s518_cmd));
		if (rc != 0)
			return rc;
		dprintk(4, "writing address %x with value %x\n", fh->io, p.val);
		s518_mutex_lock(&dev->reset_lock);
		s518_outb(dev, p.val, fh->io, 1);
		s518_mutex_unlock(&dev->reset_lock);
		dprintk(4, "s518 ioc write port %x\n", p.val);
		break;	
	}
	case S518_IOC_READ_BYTE:
	{
		struct s518_cmd p;
		dprintk(8, "read port\n");
		rc = copy_from_user(&p, argp, sizeof(struct s518_cmd));
		if (rc != 0) {
			printk("failed copy from user\n");
			return rc;
		}
		s518_mutex_lock(&dev->reset_lock);
		p.val = s518_inb(dev, fh->io);
		s518_mutex_unlock(&dev->reset_lock);
		dprintk(4, "s518 ioc read port[%x]: %x\n", fh->io, p.val);
		rc = copy_to_user(argp, &p, sizeof(struct s518_cmd));
		break;
	}
	case S518_IOC_READ_STATUS:
	{
		struct s518_cmd p;
		dprintk(4, "read port\n");
		rc = copy_from_user(&p, argp, sizeof(struct s518_cmd));
		if (rc != 0) {
			printk("failed copy from user\n");
			return rc;
		}
		s518_mutex_lock(&dev->reset_lock);
		p.val = s518_inb(dev, fh->io + 1);
		if (p.val & 0x10) {
			printk("firmware fault?\n");
			dev->fault_flags |= 0x10;
		}
		s518_mutex_unlock(&dev->reset_lock);
		dprintk(4, "s518 ioc read status[%x]: %x\n", fh->io, p.val);
		rc = copy_to_user(argp, &p, sizeof(struct s518_cmd));
		break;
	}    
	case S518_IOC_READ_MULTIPLE:
	{
		struct s518_read_mult p;
		int j;
		rc = copy_from_user(&p, argp, sizeof(struct s518_read_mult));
		if (rc != 0) {
			printk("failed copy from user\n");
			return rc;
		}
		if (p.nbytes > MAX_READ_BYTES) {
			printk("too many bytes requested\n");
			return -EINVAL;
		}
		for (j = 0; j < p.nbytes; j++) {
			p.val[j] = s518_inb(dev, fh->io);
			dprintk(4, "s518 ioc read port[%x]: %x\n", fh->io, p.val[j]);
		}
		rc = copy_to_user(argp, &p, sizeof(struct s518_read_mult));
		break;
	}
	case S518_IOC_WRITE_COMMAND:
	{
		struct s518_cmd_full p;
		rc = copy_from_user(&p, argp, sizeof(struct s518_cmd_full));
		if (rc != 0)
			return rc;
		s518_mutex_lock(&dev->reset_lock);
		rc = send_command(dev, p.cmd, p.len);
		s518_mutex_unlock(&dev->reset_lock);
		dprintk(4, "s518 ioc write command %x %x %x, %d\n", p.cmd[0], 
			p.cmd[1], p.cmd[2], p.len);
		break;	
	}    
	case S518_IOC_RESET_BOARD:
	{
		int j;
		int status;
		j = 0;
#define MAX_RESET_RETRIES 3
		s518_mutex_lock(&dev->reset_lock);
		dev->in_reset = 1;
		while (1) {
			s518_outb(dev, 0, fh->io + 1, 1);
			status = s518_inb(dev, fh->io + 1);
			if (status & STAT_FAULT) {
				rc = handshakewait(dev, STAT_FAULT, STAT_FAULT, 1000);
				if (rc != 0) {
					dev->fault_flags |= FAULT_LED;
				} else {
					status = s518_inb(dev, fh->io + 1);
					if ((status & STAT_DAV) != 0) {
						// fault_dav
						dev->fault_flags |= FAULT_DAV;
					}
					if ((status & STAT_CRMT) == 0) {
						// fault CRMT
						dev->fault_flags |= FAULT_CRMT;
					}
				}
				break;            
			} else if (j++ > MAX_RESET_RETRIES) {
				dev->fault_flags |= FAULT_LED;
				printk("failed to reset board!\n");
				dev->in_reset = 0;
				s518_mutex_unlock(&dev->reset_lock);

				return -EFAULT;
			}
		}
		dev->in_reset = 0;
		s518_mutex_unlock(&dev->reset_lock);
		init_all_channels(dev);
		printk("reset board. fault flags cleared\n");
		dev->board_state = 0;
		dev->fault_flags = 0;
		dev->is_mode_high_speed = 0;
		dev->is_mode_reject_50hz = 0;
		printk("reset board\n");
		break;
	}
	case S518_IOC_GET_CHANNEL:
	{
		struct s518_channel_prop p;
		int idx;
		struct s518_channel *ch;
		rc = copy_from_user(&p, argp, sizeof(struct s518_channel_prop));
		if (rc != 0)
			return rc;
		idx = p.idx;
		if (idx < 0)
			return -EINVAL;
		if (idx >= NUM_518_CHANNELS)
			return -EINVAL;
		ch = &dev->ch[idx];
		//printk("idx %p %x\n", ch, idx);
		p.lsdc = ch->lsdc;
		p.HiLim = ch->HiLim;
		p.LoLim = ch->LoLim;
		p.LoLimEnable = ch->LoLimEnable;
		p.HiLimEnable = ch->HiLimEnable;
		p.LoAlarm = ch->LoAlarm;
		p.HiAlarm = ch->HiAlarm;
		p.FailHigh = ch->FailHigh;
		p.Filter = ch->Filter;
		p.GagemVperV = ch->GagemVperV;
		p.GageMaxLoad = ch->GageMaxLoad;
		//printk("p.lsdc %d\n", p.lsdc);
		rc = copy_to_user(argp, &p, sizeof(struct s518_channel_prop));
		break;
	}
	case S518_IOC_GET_BOARD:
	{
		struct s518_board_prop p;
		p.fault_flags = dev->fault_flags;
		p.board_state = dev->board_state;
		p.is_mode_high_speed = dev->is_mode_high_speed;
		p.is_mode_reject_50hz = dev->is_mode_reject_50hz;
		p.num_chans = NUM_518_CHANNELS;
		rc = copy_to_user(argp, &p, sizeof(struct s518_board_prop));
		break;
	}    
	case S518_IOC_GET_FAULTFLAGS:
	{
		rc = copy_to_user(argp, &dev->fault_flags, sizeof(int));
		break;
	}    
	default:

		printk("s518: unknown ioctl\n");
		return -EINVAL;
	}
	return rc;
}

struct s518_dev *s518_devdata(struct file *file)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
	return &s518_device[iminor(file->f_dentry->d_inode)];
#else
	return &s518_device[iminor(file->f_path.dentry->d_inode)];
#endif
}

static int s518_open( struct inode *inode, struct file *file )
{
	struct s518_dev *dev = s518_devdata(file);
	struct s518_fh *fh = NULL;

	s518_mutex_lock(&dev->open_lock);
	fh = kmalloc(sizeof(struct s518_fh), GFP_KERNEL);
	if (fh == NULL) {
		s518_mutex_unlock(&dev->open_lock);
		return -ENOMEM;
	}
	fh->dev = dev;
	fh->io = dev->io;
	file->private_data = fh;
	dev->users++;
	s518_mutex_unlock(&dev->open_lock);
	dprintk(4, "%s\n", __func__);
	return 0;
}

static int s518_close( struct inode *inode, struct file *file )
{
	struct s518_dev *dev;
	struct s518_fh *fh = file->private_data;

	dev = fh->dev;
	s518_mutex_lock(&dev->open_lock);
	kfree(fh);
	dev->users--;
	s518_mutex_unlock(&dev->open_lock);
		
	dprintk(4, "%s\n", __func__);
	return 0;
}

/* file operations */
struct file_operations s518_fops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION_NOINODE
        unlocked_ioctl: s518_ioctl,
#else
        ioctl: s518_ioctl,
#endif
        open: s518_open,
        release: s518_close
};

int init_module( void )
{
	int result;
	int i;
	printk("\nInstalling 518 driver...");
	/* Count the number of 518s */
	for(num_cards = 0; io[num_cards] != 0 && num_cards < MAX_CARDS; num_cards++);
    
	/* if there are no 518s quit */
	if(num_cards == 0) {
		printk ("Missing 'io=' module parameter.\n");
		result = -1;
		goto fail_io;
	}
	printk("for %d cards\n", num_cards);
	s518_major = register_chrdev(S518_MAJOR, "s518", &s518_fops);
	/* Register the device with the kernel and get a major number. */
	if (s518_major < 0) {
		printk( "Unable to get major number %i for 518.\n", S518_MAJOR );
		goto fail_register;
	}
	/* create class for udev */
	s518_class = class_create(THIS_MODULE, "s518");
	if (IS_ERR(s518_class)) {
		goto fail_class;
	}
    
	for (i = 0; i < num_cards; i++) {
		printk("518(%d) io=%#x\n", i, (unsigned int)io[i]);
		/* create the device */
		device_create(s518_class, NULL, MKDEV(S518_MAJOR, i), NULL,
			      "%s%d", "s518_", i);
		/* request the region */
		if (request_region( io[i], ADDRESS_SPACE, "s518") == NULL) {
			printk( "518(%d) I/O address space is already occupied: %#lx.\n", i, io[i]);
			goto fail_region;
		}
		s518_mutex_init(&s518_device[i].open_lock);
		s518_mutex_init(&s518_device[i].reset_lock);
		s518_device[i].io = io[i];
		s518_device[i].users = 0;
		s518_device[i].minor = i;
		init_all_channels(&s518_device[i]);
	}
	printk( "518 driver successfully installed.\n" );
	return 0;
fail_init:
fail_region:
	for (i = 0; i < num_cards; i++) {	
		device_destroy(s518_class, MKDEV(S518_MAJOR, i));
	}
fail_device:
	class_destroy(s518_class);
fail_class:
	unregister_chrdev( S518_MAJOR, "s518" );
fail_register:
fail_io:
	printk( "518 driver installation failed!\n" );
	return result;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// MODULE SHUTDOWN.  This is called when the module is unloaded.

void cleanup_module( void )
{
	int i;
	for (i = 0; i < num_cards; i++) {
		// Release the I/O space occupied by the hardware.
		release_region( io[i], ADDRESS_SPACE );
		device_destroy(s518_class, MKDEV(S518_MAJOR, i));
	}
	// Unregister the driver with the kernel.
	class_destroy(s518_class);
	unregister_chrdev( S518_MAJOR, "s518" );
	printk( "518 driver removed.\n" );
}


